from __future__ import print_function
from urlparse import urlparse
import shutil
from subprocess import (
    STDOUT,
    check_output,
    CalledProcessError,
    check_call,
)
import os
import re
import logging
import urllib2

import fileutils


class CommandFailure(Exception):
    def __init__(self, message, original_exception):
        self.message = message
        self.original_exception = original_exception


class NotABranch(Exception):
        pass


class NotSameBranch(Exception):
    pass


class BranchDiverged(Exception):
        pass


def log_failure(cmd, message, fatal=False):
    try:
        logging.info(message)
        check_output(cmd, stderr=STDOUT)
        return True
    except CalledProcessError as e:
        logging.error(e.output)
        if fatal:
            raise CommandFailure(e.output, e)
        else:
            return False
    except OSError as e:
        logging.error(e.message)
        if fatal:
            raise CommandFailure(e.message, e)
        else:
            return False


def strip_trailing_slash(value):
    if value[-1] == "/":
        return value[:-1]
    return value


class SourceHandler(object):
    schemes = tuple()

    def __init__(self, source):
        self.source = source


class BzrSourceHandler(SourceHandler):
    """Check out a bazaar working tree"""

    schemes = (
        "bzr",
        "bzr+ssh",
        "lp",
        "bzr+http",
        "bzr+https",
        "nosmart+bzr",
        "nosmart+bzr+ssh",
        "nosmart+lp",
    )

    def __init__(self, source):
        super(BzrSourceHandler, self).__init__(source)
        scheme = urlparse(source).scheme
        if scheme in ("bzr+http", "bzr+https"):
            # Strip 'bzr+'.
            self.source = source[4:]
        else:
            self.source = source
        self.dest_source = "unset"

    def has_lp_auth(self):
        return os.path.exists(os.path.join(os.environ['HOME'],
                                           '.bazaar/authentication.conf'))

    def checkout_branch(self, dest, options=None):
        if options is None:
            options = {}
        parent_dir = os.path.dirname(dest)
        if parent_dir and not os.path.exists(parent_dir):
            os.makedirs(parent_dir)
        lightweight = options.get('lightweight', '')
        cmd = ['bzr']
        if lightweight.lower() in ('true', 't', '1'):
            cmd.append('checkout')
            cmd.append('--lightweight')
            msg = "Checking out {} to {}"
        else:
            cmd.append('branch')
            msg = "Branching {} to {}"
        if options.get('revno'):
            cmd.append('-r')
            cmd.append(options.get('revno'))
        cmd = cmd + [self.source, dest]
        return log_failure(cmd, msg.format(self.source, dest))

    def update_branch(self, dest):
        cmd = ['bzr', 'update', dest]
        return log_failure(cmd, "No revno specified: Updating {} to tip".format(dest, self.source))

    def pull_branch(self, dest, options=None):
        if options is None:
            options = {}
        cmd = ["bzr", "pull"]
        if options.get('overwrite'):
            cmd.append('--overwrite')
        if options.get('revno'):
            cmd.append('-r')
            cmd.append(options.get('revno'))
        cmd = cmd + ["-d", dest]
        return log_failure(cmd, "Pulling {} from parent ({})".format(dest, self.source))

    def revno_branch(self, dest, revno):
        cmd = ['bzr', 'update', '-r', revno, dest]
        return log_failure(cmd, "Checking out revision {} of {}".format(revno, self.source))

    def parent_branch(self, branch, local=False):
        bzr_cmd = ("bzr", "info", branch)
        output = check_output(bzr_cmd)
        # 'repository branch' has higher relevance than 'branch root', e.g:
        # m_repo:   bzr info lp:charms/trusty/ceph
        # m_root:   bzr info lp:mojo
        # m_parent: bzr info /path/to/local/branch
        m_repo = m_root = m_parent = None
        if local:
            m_parent = re.match(r".*parent branch: (\S+)", output, re.MULTILINE | re.DOTALL)
        else:
            m_repo = re.match(r".*repository branch: (\S+)", output, re.MULTILINE | re.DOTALL)
            m_root = re.match(r".*branch root: (\S+)", output, re.MULTILINE | re.DOTALL)
        match = m_repo or m_root or m_parent
        if match:
            parent = match.group(1)
            parent = strip_trailing_slash(parent)
            return parent
        else:
            return None

    def normalize_lp_branch(self, branch):
        if branch.startswith(('lp:', 'nosmart+lp:')):
            if self.has_lp_auth():
                if '~' not in branch:
                    branch = branch.replace('lp:', 'lp:+branch/')
                branch = branch.replace('lp:', 'bzr+ssh://bazaar.launchpad.net/')
            else:
                # properly map lp: to http://bazaar.launchpad.net/...
                # for anon branching
                branch = self.parent_branch(branch)
        return branch

    def is_same_branch(self, dest):
        self.source = strip_trailing_slash(self.source).strip()
        self.source = self.normalize_lp_branch(self.source)
        self.dest_source = self.parent_branch(dest, True)
        return self.dest_source == self.source

    def is_bzr_branch(self, branch):
        branch = self.normalize_lp_branch(branch)
        bzr_cmd = ("bzr", "revno", branch)
        devnull = open('/dev/null', 'w')
        try:
            check_call(bzr_cmd, stdout=devnull)
            return True
        except CalledProcessError as e:
            if e.returncode == 3:
                return False
            else:
                raise e

    def get_branch_revno(self, branch):
        """Return the branch revno"""
        cmd = ['bzr', 'revno', branch]
        return check_output(cmd).strip()

    def get_working_tree_revno(self, branch):
        """Return the working tree revno"""
        cmd = ['bzr', 'version-info', branch]
        output = check_output(cmd).split('\n')
        for line in output:
            if line.startswith('revno:'):
                line = line.strip()
                return line.split()[1]

    def working_tree_matches_revno(self, branch):
        """Compare working tree revno against branch revno
           return True or False"""
        return self.get_branch_revno(branch) == self.get_working_tree_revno(branch)

    def get(self, dest, options=None):
        if not options:
            options = {}
        if not self.is_bzr_branch(self.source):
            raise NotABranch("{} is not a bzr branch. Is it a private branch? Check permissions on the branch.".format(self.source))
        if os.path.exists(dest):
            if not self.is_bzr_branch(dest):
                raise NotABranch("{} is not a bzr branch, it may be an empty directory".format(dest))
                return False
            # if the parent is the same, update the branch
            if self.is_same_branch(dest):
                # Pull from parent
                if not self.pull_branch(dest, options):
                    raise BranchDiverged("{} failed: {} and {} have diverged".format(
                                         dest, dest, self.source))
                    return False
                # Handle downgrading
                if options.get("revno") and not options.get("overwrite"):
                    # Update only if working tree does not match
                    # requested revno
                    if not self.get_working_tree_revno(dest) == options.get("revno"):
                        if not self.revno_branch(dest, options.get("revno")):
                            return False
                # Handle upgrade from previously downgraded
                # working tree not at tip on disk
                if not options.get("revno"):
                    # Update only if working tree is not at tip
                    if not self.working_tree_matches_revno(dest):
                        if not self.update_branch(dest):
                            return False
            elif options.get("overwrite"):
                logging.info("Overwriting {}".format(dest))
                shutil.rmtree(dest)
                if not self.checkout_branch(dest, options):
                    return False
            else:
                self.is_same_branch(dest)
                raise NotSameBranch(
                    "{} failed: {}'s parent is {} not {}. They are not the "
                    "same branch.".format(dest, dest, self.dest_source,
                                         self.source))
        else:
            if not self.checkout_branch(dest, options):
                return False

        logging.info("{} is at revno {}".format(dest, self.get_working_tree_revno(dest)))

        return True


class HttpFileHandler(SourceHandler):
    """Download plain files via http(s)"""

    schemes = (
        "http",
        "https",
    )

    def get(self, dest, options=None):
        if not options:
            options = {}
        if os.path.exists(dest):
            if options.get("overwrite"):
                os.unlink(dest)
            else:
                logging.info("Skipping existing dest {}".format(dest))
                return True
        logging.info("Downloading {} to {}".format(self.source, dest))
        try:
            response = urllib2.urlopen(self.source)
        except urllib2.URLError as e:
            logging.error("Failed to download {}: {}".format(self.source, e.reason))
            return False
        with open(dest, "w") as f:
            f.write(response.read())
        return True


class LocalHandler(SourceHandler):
    """Copy local files. The special source '@' indicates that the destination
    is a directory."""

    schemes = (
        '',
        'file',
    )

    def get(self, dest, options=None):
        if not options:
            options = {}

        if self.source == "@":
            logging.info("Creating directory {}".format(dest))
            fileutils.mkdir(dest, overwrite=options.get("overwrite", False))
            return True

        method = options.get("method", "copy")
        if method == "copy":
            logging.info("Copying {} to {}".format(self.source, dest))
            fileutils.copy(self.source, dest)
        elif method == "rsync":
            logging.info("Rsyncing {} to {}".format(self.source, dest))
            fileutils.rsync(self.source, dest)
        elif method == "link":
            logging.info("Creating symbolic link {} to {}".format(dest, self.source))
            fileutils.link(self.source, dest, overwrite=options.get(
                "overwrite", True))
        elif method == "hardlink":
            logging.info("Creating hard link {} to {}".format(dest, self.source))
            fileutils.link(self.source, dest, symbolic=False)

        return True


class DuplicateHandlerError(Exception):
    pass


def handler_types(handler_class=None):
    "Build a registry of handlers"
    if handler_class is None:
        handler_class = SourceHandler
    types = {}
    for subclass in handler_class.__subclasses__():
        for scheme in subclass.schemes:
            if scheme in types:
                new_handler = subclass.__name__
                old_handler = types[scheme].__name__
                raise DuplicateHandlerError("{} and {}".format(old_handler, new_handler))
            else:
                types[scheme] = subclass
    return types


class NoSuchHandlerError(Exception):
    pass


def handler_for_url(url):
    "Build a handler based on the given URL"
    if getattr(handler_types, "cache", None) is None:
        handler_types.cache = handler_types()
    scheme = urlparse(url).scheme
    if scheme in handler_types.cache:
        return handler_types.cache[scheme](url)
    raise NoSuchHandlerError("No handler found for URL: {}".format(url))
