import os
from unittest import TestCase
from urlparse import urlparse
from tempfile import mkdtemp
from urllib2 import URLError
import subprocess
import shutil
try:
    from cStringIO import StringIO
except:
    from StringIO import StringIO

from mock import (
    patch,
    MagicMock,
    mock_open,
)

from codetree.handlers import (
    strip_trailing_slash,
    log_failure,
    CommandFailure,
    SourceHandler,
    BzrSourceHandler,
    LocalHandler,
    HttpFileHandler,
    NotSameBranch,
)

BzrURLs = (
    "bzr://example.com/foo/",
    "lp:debian/apt/",
    "bzr+ssh://bazaar.launchpad.net/~foo/bar/trunk/",
    "bzr+http://example.com/foo/",
    "bzr+https://example.com/foo/",
    "lp:debian/apt",
)

HttpURLs = (
    "http://example.com/foo.txt",
    "https://example.com/foo.txt",
)

LocalURLs = (
    "file:///etc/hosts",
    "/etc/hosts",
    "etc/hosts",
    "../etc/hosts",
    "hosts,"
    "/etc",
)

BzrInfoOutputs = {
    'anon': {
        'bzrinfo': """Repository branch (format: 2a)
Location:
  shared repository: http://bazaar.launchpad.net/~openstack-charmers/charms/trusty/ceph/trunk/
  repository branch: http://bazaar.launchpad.net/~openstack-charmers/charms/trusty/ceph/trunk/

Related branches:
  parent branch: http://bazaar.launchpad.net/~openstack-charmers/charms/precise/ceph/trunk/
     stacked on: /+branch-id/905249
""",
        'parent': 'http://bazaar.launchpad.net/~openstack-charmers/charms/trusty/ceph/trunk',
    },
    'auth': {
        'bzrinfo': """Repository branch (format: unnamed)
Location:
  shared repository: bzr+ssh://bazaar.launchpad.net/+branch/charms/trusty/ceph/
  repository branch: bzr+ssh://bazaar.launchpad.net/+branch/charms/trusty/ceph/

Related branches:
  parent branch: bzr+ssh://bazaar.launchpad.net/+branch/precise/ceph/trunk/
     stacked on: /+branch-id/905249
""",
        'parent': 'bzr+ssh://bazaar.launchpad.net/+branch/charms/trusty/ceph',
    },
    'local': {
        'bzrinfo':  """Standalone tree (format: 2a)
Location:
  branch root: .

Related branches:
  parent branch: bzr+ssh://bazaar.launchpad.net/+branch/codetree/
""",
        'parent': 'bzr+ssh://bazaar.launchpad.net/+branch/codetree',
    },
}


def was_called_with_cmd(mock, cmd):
    for call_args in mock.call_args_list:
        if call_args[0][0] == cmd:
            return True
    return False


def shellcmd(cmd):
    try:
        subprocess.check_output(cmd, shell=True, stderr=subprocess.STDOUT)
    except subprocess.CalledProcessError as e:
        print "------- output --------"
        print e.output
        print "-----------------------"
        raise


class TestLogHandlers(TestCase):
    @patch("codetree.handlers.check_output")
    def test_log_failure_success(self, _call):
        cmd = ("echo", "hello world")
        log_failure(cmd, "Saying hello")
        assert(was_called_with_cmd(_call, cmd))

    def test_log_failure_failure(self):
        # Throws OSError
        cmd = ("/invalid/cmd", "hello world")
        self.assertFalse(log_failure(cmd, "Saying hello"))
        with self.assertRaises(CommandFailure):
            log_failure(cmd, "Failing", fatal=True)

        cmd = ("false",)
        self.assertFalse(log_failure(cmd, "Failing"))
        with self.assertRaises(CommandFailure):
            log_failure(cmd, "Failing", fatal=True)

    def test_strip_trailing_slash(self):
        value = "abc"
        self.assertEqual(value, strip_trailing_slash(value))
        self.assertEqual(value, strip_trailing_slash(value + "/"))


class TestSourceHandler(TestCase):
    def setUp(self):
        super(TestSourceHandler, self).setUp()
        self.source = "foo"
        self.sh = SourceHandler("foo")

    def test_stores_source(self):
        self.assertEqual(self.sh.source, self.source)

    def test_handles_nothing(self):
        self.assertEqual(SourceHandler.schemes, tuple())


class BzrSourceHandlerTest(TestCase):
    def test_url_handling(self):
        for url in BzrURLs:
            assert(urlparse(url).scheme in BzrSourceHandler.schemes)

    def test_stores_source(self):
        url = BzrURLs[0]
        bh = BzrSourceHandler(url)
        self.assertEqual(bh.source, url)

    def test_nonstandard_schemes(self):
        http_url = "bzr+http://example.com/repo"
        bh = BzrSourceHandler(http_url)
        self.assertTrue(bh.source.startswith("http"))

        https_url = "bzr+https://example.com/repo"
        bh = BzrSourceHandler(https_url)
        self.assertTrue(bh.source.startswith("https"))

    @patch("codetree.handlers.BzrSourceHandler.has_lp_auth", return_value=True)
    @patch("codetree.handlers.check_output")
    @patch("codetree.handlers.check_call", return_value=[False, True])
    @patch("codetree.handlers.logging")
    @patch("codetree.handlers.os.path.exists", return_value=True)
    @patch("codetree.handlers.shutil.rmtree")
    def test_overwite(self, _rmtree, _exists, _log, _call, _output, _auth):
        source = BzrURLs[0]
        dest = "foo"
        bh = BzrSourceHandler(source)
        bh.is_same_branch = MagicMock(return_value=False)

        # overwrite (delete) existing when asked
        options = {"overwrite": True}
        self.assertTrue(bh.get(dest, options))
        _rmtree.assert_called_with(dest)

        bh.checkout_branch = MagicMock(return_value=False)
        self.assertFalse(bh.get(dest, options))

        # don't overwrite if not asked
        options = {"overwrite": False}
        with self.assertRaises(NotSameBranch) as ctx:
            bh.get(dest, options)
        self.assertEqual(
            "foo failed: foo's parent is unset not bzr://example.com/foo/. "
            "They are not the same branch.",
            ctx.exception.message)
        _rmtree.assert_not_called()

        # don't overwrite if source = parent
        options = {"overwrite": True}
        bh.is_same_branch = MagicMock(return_value=True)
        self.assertTrue(bh.get(dest, options))
        _rmtree.assert_not_called()

    @patch("codetree.handlers.BzrSourceHandler.has_lp_auth", return_value=True)
    @patch("codetree.handlers.check_output")
    @patch("codetree.handlers.check_call", return_value=True)
    @patch("codetree.handlers.os.path.exists", return_value=False)
    def test_branches_new(self, _exists, _call, _output, _auth):
        source = BzrURLs[0]
        dest = "foo"
        bh = BzrSourceHandler(source)
        self.assertTrue(bh.get(dest))
        self.assertTrue(
            was_called_with_cmd(_output, ['bzr', 'branch', source, dest]))
        bh.checkout_branch = MagicMock(return_value=False)
        self.assertFalse(bh.get(dest))

    @patch("codetree.handlers.BzrSourceHandler.has_lp_auth", return_value=True)
    @patch("codetree.handlers.check_output")
    @patch("codetree.handlers.check_call", return_value=True)
    @patch("codetree.handlers.os.path.exists", return_value=False)
    def test_branches_new_lightweight(self, _exists, _call, _output, _auth):
        source = "lp:debian/apt"
        dest = "foo"
        bh = BzrSourceHandler(source)
        _output.return_value = ""
        options = {'lightweight': 'true'}
        self.assertTrue(bh.get(dest, options))
        self.assertTrue(
            was_called_with_cmd(
                _output, ['bzr', 'checkout', '--lightweight',
                          'lp:debian/apt', dest]))
        bh.checkout_branch = MagicMock(return_value=False)
        self.assertFalse(bh.get(dest))

    @patch("codetree.handlers.check_output")
    @patch("codetree.handlers.check_call", return_value=True)
    @patch("codetree.handlers.os.makedirs")
    def test_creates_new_dirs(self, _makedirs, _call, _output):
        source = BzrURLs[0]
        dest = "foo/bar/baz"
        bh = BzrSourceHandler(source)
        self.assertTrue(bh.get(dest))
        _makedirs.assert_called_with(os.path.dirname(dest))

    @patch("codetree.handlers.BzrSourceHandler.has_lp_auth", return_value=True)
    @patch("codetree.handlers.check_output")
    @patch("codetree.handlers.check_call", return_value=True)
    @patch("codetree.handlers.os.path.exists", return_value=True)
    def test_updates_existing(self, _exists, _call, _output, _auth):
        source = BzrURLs[0]
        dest = "foo"
        bh = BzrSourceHandler(source)
        bh.is_same_branch = MagicMock(return_value=True)
        self.assertTrue(bh.get(dest))
        assert(was_called_with_cmd(_output, ['bzr', 'pull', '-d', dest]))

        bh.update_branch = MagicMock(return_value=False)
        self.assertFalse(bh.get(dest))

    @patch("codetree.handlers.check_output")
    @patch("codetree.handlers.check_call", return_value=True)
    def test_gets_revno(self, _call, _output):
        source = BzrURLs[0]
        dest = "foo"
        bh = BzrSourceHandler(source)
        bh.is_same_branch = MagicMock()

        revno = "1"
        options = {"revno": "1"}
        self.assertTrue(bh.get(dest, options))
        assert(was_called_with_cmd(
            _output, ['bzr', 'branch', '-r', revno, source, dest]))

    @patch("codetree.handlers.BzrSourceHandler.has_lp_auth", return_value=True)
    @patch('codetree.handlers.BzrSourceHandler.get_working_tree_revno', return_value='5')
    @patch("codetree.handlers.check_output")
    @patch("codetree.handlers.check_call", return_value=True)
    @patch('codetree.handlers.os.path.exists', return_value=True)
    def test_existing_revno_no_overwrite_no_match(self, _exists, _call, _output, _revno, _auth):
        source = BzrURLs[0]
        dest = "foo"
        bh = BzrSourceHandler(source)
        bh.is_same_branch = MagicMock()

        revno = "1"
        options = {"revno": "1"}
        self.assertTrue(bh.get(dest, options))
        assert(was_called_with_cmd(
            _output, ['bzr', 'pull', '-r', revno, '-d', dest]))

        assert(was_called_with_cmd(
            _output, ['bzr', 'update', '-r', revno, dest]))

    @patch("codetree.handlers.BzrSourceHandler.has_lp_auth", return_value=True)
    @patch('codetree.handlers.BzrSourceHandler.update_branch')
    @patch('codetree.handlers.BzrSourceHandler.get_working_tree_revno', return_value='1')
    @patch("codetree.handlers.check_output")
    @patch("codetree.handlers.check_call", return_value=True)
    @patch('codetree.handlers.os.path.exists', return_value=True)
    def test_existing_revno_no_overwrite_with_match(self, _exists, _call, _output, _revno, _update, _auth):
        source = BzrURLs[0]
        dest = "foo"
        bh = BzrSourceHandler(source)
        bh.is_same_branch = MagicMock()

        revno = "1"
        options = {"revno": "1"}
        self.assertTrue(bh.get(dest, options))
        assert(was_called_with_cmd(
            _output, ['bzr', 'pull', '-r', revno, '-d', dest]))

        _update.assert_not_called()

    @patch("codetree.handlers.BzrSourceHandler.has_lp_auth", return_value=True)
    @patch('codetree.handlers.BzrSourceHandler.update_branch')
    @patch('codetree.handlers.BzrSourceHandler.get_working_tree_revno', return_value='1')
    @patch("codetree.handlers.check_output")
    @patch("codetree.handlers.check_call", return_value=True)
    @patch('codetree.handlers.os.path.exists', return_value=True)
    def test_existing_revno_with_overwrite(self, _exists, _call, _output, _update, _revno, _auth):
        source = BzrURLs[0]
        dest = "foo"
        bh = BzrSourceHandler(source)
        bh.is_same_branch = MagicMock()

        revno = "1"
        options = {"revno": "1", "overwrite": True}
        self.assertTrue(bh.get(dest, options))
        assert(was_called_with_cmd(
            _output, ['bzr', 'pull', '--overwrite', '-r', revno, '-d', dest]))

        _update.assert_not_called()

    @patch("codetree.handlers.BzrSourceHandler.has_lp_auth", return_value=True)
    @patch('codetree.handlers.BzrSourceHandler.update_branch')
    @patch('codetree.handlers.BzrSourceHandler.working_tree_matches_revno', return_value=True)
    @patch("codetree.handlers.check_output")
    @patch("codetree.handlers.check_call", return_value=True)
    @patch('codetree.handlers.os.path.exists', return_value=True)
    def test_existing_no_revno_with_match(self, _exists, _call, _output, _tree, _update, _auth):
        source = BzrURLs[0]
        dest = "foo"
        bh = BzrSourceHandler(source)
        bh.is_same_branch = MagicMock()

        options = {}
        self.assertTrue(bh.get(dest, options))
        assert(was_called_with_cmd(
            _output, ['bzr', 'pull', '-d', dest]))

        _update.assert_not_called()

    @patch("codetree.handlers.BzrSourceHandler.has_lp_auth", return_value=True)
    @patch('codetree.handlers.BzrSourceHandler.working_tree_matches_revno', return_value=False)
    @patch("codetree.handlers.check_output")
    @patch("codetree.handlers.check_call", return_value=True)
    @patch('codetree.handlers.os.path.exists', return_value=True)
    def test_existing_no_revno_no_match(self, _exists, _call, _output, _tree, _auth):
        source = BzrURLs[0]
        dest = "foo"
        bh = BzrSourceHandler(source)
        bh.is_same_branch = MagicMock()

        options = {}
        self.assertTrue(bh.get(dest, options))
        assert(was_called_with_cmd(
            _output, ['bzr', 'pull', '-d', dest]))

        assert(was_called_with_cmd(
            _output, ['bzr', 'update', dest]))

    def test_same_branch(self):
        parent = mkdtemp()
        self.addCleanup(shutil.rmtree, parent)
        shellcmd("bzr init {}".format(parent))
        bh = BzrSourceHandler(parent)

        child_tmp = mkdtemp()
        self.addCleanup(shutil.rmtree, child_tmp)

        # is same
        child = os.path.join(child_tmp, "child") + "/"
        shellcmd("bzr branch {} {}".format(parent, child))
        self.assertTrue(bh.is_same_branch(child))

        # is not the same
        nonchild = os.path.join(child_tmp, "nonchild")
        shellcmd("bzr branch {} {}".format(child, nonchild))
        self.assertFalse(bh.is_same_branch(nonchild), "test: is not same")

        # is standalone
        stdalone = os.path.join(child_tmp, "stdalone")
        shellcmd("bzr init {}".format(stdalone))
        self.assertFalse(bh.is_same_branch(stdalone), "test: is standalone")

    @patch('codetree.handlers.check_output')
    def test_parent_branch(self, _check_output):
        lh = BzrSourceHandler("/some/file")
        for key in BzrInfoOutputs:
            bzr = BzrInfoOutputs[key]
            _check_output.return_value = bzr['bzrinfo']
            parent = lh.parent_branch("foo", key == 'local')
            _check_output.assert_called_with(('bzr', 'info', 'foo'))
            self.assertEqual(parent, bzr['parent'])
            _check_output.reset_mock()

    @patch('codetree.handlers.BzrSourceHandler.parent_branch')
    @patch('codetree.handlers.BzrSourceHandler.has_lp_auth', return_value=True)
    def test_normalize_lp_branch_auth(self, _has_lp_auth, _parent_branch):
        lh = BzrSourceHandler("/some/file")
        # ~user, only replaces lp: by bzr+ssh://...
        normalized = lh.normalize_lp_branch("lp:~bob/foo")
        _parent_branch.assert_not_called()
        self.assertEqual(normalized,
                         'bzr+ssh://bazaar.launchpad.net/~bob/foo')
        # non ~user, adds +branch
        normalized = lh.normalize_lp_branch("lp:foo")
        self.assertEqual(normalized,
                         'bzr+ssh://bazaar.launchpad.net/+branch/foo')
        _parent_branch.assert_not_called()

    @patch('codetree.handlers.BzrSourceHandler.parent_branch')
    @patch('codetree.handlers.BzrSourceHandler.has_lp_auth', return_value=False)
    def test_normalize_lp_branch_anon(self, _has_lp_auth, _parent_branch):
        _parent_branch.return_value = 'http://foo/bar'
        lh = BzrSourceHandler("/some/file")
        # normalize doesn't change LP paths if it's anon LP,
        # but uses parent_branch() value
        normalized = lh.normalize_lp_branch("lp:foo")
        _parent_branch.assert_called_with("lp:foo")
        self.assertEqual(normalized, 'http://foo/bar')


class TestLocalHandler(TestCase):
    def test_url_handling(self):
        for local_url in LocalURLs:
            assert(urlparse(local_url).scheme in LocalHandler.schemes)

    @patch("codetree.handlers.fileutils.mkdir")
    def test_creates_directory(self, _mkdir):
        lh = LocalHandler("@")
        self.assertTrue(lh.get("foo"))
        _mkdir.assert_called_with('foo', overwrite=False)

    @patch("codetree.handlers.fileutils.copy")
    def test_copies_files_dirs(self, _copy):
        lh = LocalHandler("/some/file")
        options = {"method": "copy"}
        self.assertTrue(lh.get("foo", options))
        _copy.assert_called_with("/some/file", "foo")

    @patch("codetree.handlers.fileutils.rsync")
    def test_rsyncs_files_dirs(self, _rsync):
        lh = LocalHandler("/some/file")
        options = {"method": "rsync"}
        self.assertTrue(lh.get("foo", options))
        _rsync.assert_called_with("/some/file", "foo")

    @patch("codetree.handlers.fileutils.link")
    def test_links_files_dirs(self, _link):
        lh = LocalHandler("/some/file")
        # symbolic link
        options = {"method": "link"}
        self.assertTrue(lh.get("foo", options))
        _link.assert_called_with("/some/file", "foo", overwrite=True)
        # hard link
        options = {"method": "hardlink"}
        self.assertTrue(lh.get("foo", options))
        _link.assert_called_with("/some/file", "foo", symbolic=False)

    @patch("codetree.handlers.fileutils.copy")
    def test_copy_is_default(self, _copy):
        lh = LocalHandler("/some/file")
        self.assertTrue(lh.get("foo"))
        _copy.assert_called_with("/some/file", "foo")


class TestHttpFileHandler(TestCase):
    def test_url_handling(self):
        for http_url in HttpURLs:
            assert(urlparse(http_url).scheme in HttpFileHandler.schemes)

    @patch('codetree.handlers.os.unlink')
    @patch('codetree.handlers.os.path.exists')
    @patch('codetree.handlers.urllib2.urlopen')
    def test_gets_file(self, _urlopen, _exists, _unlink):
        destfile = "foo"

        _urlopen.return_value = StringIO("words words")
        hh = HttpFileHandler(HttpURLs[0])

        # New file
        _open = mock_open()
        _exists.return_value = False
        with patch('codetree.handlers.open', _open, create=True):
            self.assertTrue(hh.get(destfile))
        self.assertFalse(_unlink.called)
        _open.assert_called_with(destfile, "w")
        _urlopen.assert_called_with(HttpURLs[0])

    @patch('codetree.handlers.os.unlink')
    @patch('codetree.handlers.os.path.exists')
    @patch('codetree.handlers.urllib2.urlopen')
    def test_gets_file_no_overwrite(self, _urlopen, _exists, _unlink):
        destfile = "foo"

        _urlopen.return_value = StringIO("words words")
        hh = HttpFileHandler(HttpURLs[0])

        # Existing file
        _open = mock_open()
        _exists.return_value = True
        with patch('codetree.handlers.open', _open, create=True):
            self.assertTrue(hh.get(destfile))
        self.assertFalse(_unlink.called)
        self.assertFalse(_open.called)
        self.assertFalse(_urlopen.called)

    @patch('codetree.handlers.os.unlink')
    @patch('codetree.handlers.os.path.exists')
    @patch('codetree.handlers.urllib2.urlopen')
    def test_gets_file_with_overwite(self, _urlopen, _exists, _unlink):
        destfile = "foo"

        _urlopen.return_value = StringIO("words words")
        hh = HttpFileHandler(HttpURLs[0])

        # Overwrite existing file
        _open = mock_open()
        _exists.return_value = True
        with patch('codetree.handlers.open', _open, create=True):
            self.assertTrue(hh.get(destfile, options={"overwrite": True}))
        _unlink.assert_called_with(destfile)
        _open.assert_called_with(destfile, "w")
        _urlopen.assert_called_with(HttpURLs[0])

    @patch('codetree.handlers.os.unlink')
    @patch('codetree.handlers.os.path.exists')
    @patch('codetree.handlers.urllib2.urlopen')
    def test_gets_file_bad_url(self, _urlopen, _exists, _unlink):
        destfile = "foo"

        _urlopen.return_value = StringIO("words words")
        hh = HttpFileHandler(HttpURLs[0])

        # Broken source
        _open = mock_open()
        _exists.return_value = False
        _urlopen.side_effect = URLError('failed')
        with patch('codetree.handlers.open', _open, create=True):
            self.assertFalse(hh.get(destfile))
